/*
 * hello.c
 *
 * Author: Lain Musgrove (lain.proliant@gmail.com)
 * Date: Tuesday August 29, 2023
 */

#include <stdio.h>
#include <stdlib.h>

int main() {
    printf("Hello, world!");
    return 0;
}
